/*
 * feedbackd
 *
 * Copyright (c) 2016 - 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <stdlib.h>
#include <stdbool.h>
#include <device/power.h>

#include "core/log.h"
#include "core/list.h"
#include "core/config-parser.h"
#include "haptic.h"
#include "standard-vibcore.h"

#define VIBRATION_CONF_PATH        "/usr/share/feedback/vibration.conf"

struct vibration_config {
	char *pattern;
	int pattern_duration;
	dd_list *data;
};

struct duration_data {
	int duration;
	int wait;
};

static dd_list *vib_conf_list;
static guint duration_timer;

static t_vibrate_monotone real_vibrate_monotone;

static int vibration_load_config(struct parse_result *result, void *user_data)
{
	struct vibration_config *conf;
	struct duration_data *data;
	char *value;
	char *check;
	int len;

	if (!result)
		return 0;

	if (!MATCH(result->section, "Vibration"))
		return 0;


	if (!result->name || !result->value)
		return 0;

	conf = (struct vibration_config *)calloc(1, sizeof(struct vibration_config));
	if (!conf) {
		_E("fail to alloc");
		return -ENOMEM;
	}

	conf->pattern_duration = 0;
	conf->pattern = strdup(result->name);
	if (!conf->pattern) {
		_E("fail to copy %s pattern data", result->name);
		free(conf);
		return -ENOMEM;
	}

	value = result->value;
	len = strlen(value);

	if (len == 0) {
		data = (struct duration_data *)malloc(sizeof(struct duration_data));
		if (!data) {
			_E("not enough memory");
			free(conf->pattern);
			free(conf);
			return -ENOMEM;
		}
		data->duration = 0;
		data->wait = 0;

		DD_LIST_APPEND(conf->data, data);
		DD_LIST_APPEND(vib_conf_list, conf);
		return 0;
	}

	/* value : 100D or 100D0W or 250D250W250D750W*/
	do {
		data = (struct duration_data *)malloc(sizeof(struct duration_data));
		if (!data) {
			_E("not enough memory");
			free(conf->pattern);
			free(conf);
			return -ENOMEM;
		}
		data->duration = 0;
		data->wait = 0;

		check = strchr(value, 'D');
		if (check) {
			*check = '\0';
			data->duration = strtol(value, NULL, 10);
			len = len - strlen(value) - 1;
			value = check + 1;
			conf->pattern_duration += data->duration;
		}

		check = strchr(value, 'W');
		if (check) {
			*check = '\0';
			data->wait = strtol(value, NULL, 10);
			len = len - strlen(value) - 1;
			value = check + 1;
			conf->pattern_duration += data->wait;
		}

		DD_LIST_APPEND(conf->data, data);
		if (data->duration == 0 && data->wait == 0)
			break;
	} while (len > 0);

	DD_LIST_APPEND(vib_conf_list, conf);

	return 0;
}

int standard_config_parse()
{
	return config_parse(VIBRATION_CONF_PATH, vibration_load_config, NULL);
}

int standard_is_supported(const char *pattern)
{
	dd_list *elem;
	struct vibration_config *conf;
	size_t len;
	int ret;

	if (!pattern)
		return -EINVAL;

	len = strlen(pattern) + 1;
	ret = 0;
	DD_LIST_FOREACH(vib_conf_list, elem, conf) {
		if (!conf->pattern)
			continue;
		if (!strncmp(conf->pattern, pattern, len)) {
			ret = true;
			break;
		}
	}

	return ret;
}

static gboolean haptic_duration_play(void *data)
{
	dd_list *head, *n, *next;
	struct haptic_data *h_data;
	struct duration_data *node;
	int ret = 0;

	if (duration_timer) {
		g_source_remove(duration_timer);
		duration_timer = 0;
	}

	if (!data) {
		cur_h_data.handle = INVALID_HANDLE;
		cur_h_data.priority = PRIORITY_MIN;
		goto out;
	}

	h_data = (struct haptic_data *)data;
	if (h_data->stop) {
		h_data->stop = false;
		cur_h_data.handle = INVALID_HANDLE;
		cur_h_data.priority = PRIORITY_MIN;
		goto out;
	}

	head = h_data->vibration_data;
	DD_LIST_FOREACH_SAFE(head, n, next, node) {
		_D("Play: %dms and Wait: %dms", node->duration, node->wait);
		if (!node->duration) {
			cur_h_data.handle = INVALID_HANDLE;
			cur_h_data.priority = PRIORITY_MIN;
			break;
		}

		if (next) {
			h_data->vibration_data = next;
			duration_timer = g_timeout_add((node->duration + node->wait), haptic_duration_play, (void *)h_data);
		} else
			duration_timer = g_timeout_add((node->duration + node->wait), haptic_duration_play, NULL);

		ret = real_vibrate_monotone(h_data->handle, node->duration, h_data->level, h_data->priority, NULL);
		if (!next)
			goto out;

		break;
	}
	if (ret != 0) {
		_D("auto stop vibration");
		if (h_data)
			h_data->stop = true;
	}
out:
	return G_SOURCE_REMOVE;
}

int standard_set_vib_function(t_vibrate_monotone func)
{
	real_vibrate_monotone = func;
	return 0;
}

int standard_vibrate_effect(int device_handle, const char *pattern, int feedback, int priority)
{
	dd_list *elem;
	struct vibration_config *conf;
	size_t len;
	int ret;

	if (device_handle < 0)
		return -EINVAL;

	/* Same or higher priority pattern should be played */
	if (priority < cur_h_data.priority) {
		_E("Priority of new request is lower than old request");
		return -EPERM;
	}

	len = strlen(pattern) + 1;
	DD_LIST_FOREACH(vib_conf_list, elem, conf) {
		if (!conf->pattern)
			continue;
		if (strncmp(conf->pattern, pattern, len))
			continue;

		cur_h_data.vibration_data = conf->data;
		cur_h_data.handle = device_handle;
		cur_h_data.level = feedback;
		cur_h_data.priority = priority;
		cur_h_data.stop = false;
		_D("Play %s", conf->pattern);

		if (conf->pattern_duration <= 0)
			break;

		ret = device_power_request_lock(POWER_LOCK_CPU, conf->pattern_duration);
		if (ret != DEVICE_ERROR_NONE)
			_E("Failed to request power lock");
		haptic_duration_play((void *)&cur_h_data);
		break;
	}

	return 0;
}

int standard_vibrate_close()
{
	cur_h_data.handle = INVALID_HANDLE;
	cur_h_data.priority = PRIORITY_MIN;

	if (duration_timer) {
		_I("Remove duration_timer");
		g_source_remove(duration_timer);
		duration_timer = 0;
	}

	return 0;
}

